# Copyright 2013 Jorge Estrada. 
#
# This file is part of azucar.buffer project. 
# It is subject to the license terms in the 
# LICENSE file found in the top-level directory of this distribution. 
# No part of azucar.buffer project, including this file, may be copied,
# modified, propagated, or distributed except according to the terms contained 
# in the LICENSE file.

import math
import itertools
import json
import numpy as np

import pkg_resources

default_buffer_and_species_data_description_filename = pkg_resources.resource_filename(__name__, "data/species_description.json")

class BufferType:
    def __init__(self, name, species, pKas):
        self.name = name
        """Unique name identifying this buffer type. Type: :py:func:`str`.
        """

        self.species = species
        """List of :py:class:`SpeciesType` objects, which are the species
        forming this buffer. Example: for acetate buffer, the species are
        C2H4O2 and C2H3O2-. The list is ordered from the most protonated
        species first, to the least protonated species last.
        """

        self.pKas = pKas
        """List of pKa (:py:func:`float`) values corresponding to the deprotonation of
        the species in ``self.species``. ``self.pKas[i]`` is the pKa for the
        deprotonation of species ``self.species[i]`` into ``self.species[i+1]``.
        """

class SpeciesType:
    def __init__(self, formula, charge):
        self.formula = formula
        """Unique formula identifying this molecular species. Type: :py:func:`str`.
        Example: For acetate acid, formula is C2H4O2. For acetate ion,
        formula is C2H3O2-.
        """

        self.charge = int(charge)
        "Integer charge of this molecular species."

class BufferAndSpeciesTypesDataCollection:
    def __init__(self):
        self.buffer_types = []
        "List of all the :py:class:`BufferType` included in this collection."

        self.species_types = []
        "List of all the :py:class:`SpeciesType` included in this collection."

    def get_buffer_type(self, name):
        """Get the buffer of this collection with the given name.

        :param name: Unique name that identifies the buffer type.
        :type name: :py:func:`str`

        :returns: The corresponding :py:class:`BufferType`.

        :raises: Exception if the buffer type is not in this collection.
        """
        for buffer in self.buffer_types:
            if buffer.name == name:
                return buffer

        raise Exception("Buffer not found.")
        
    def get_buffer_type_of_species(self, species):
        """Get the buffer of this collection that includes the given species.

        :param species: The species belonging to the buffer searched.
        :type species: :py:class:`SpeciesType`

        :returns: The corresponding :py:class:`BufferType`.

        :raises: Exception if the buffer type is not in this collection.
        """

        for buffer in self.buffer_types:
            if species in buffer.species:
                return buffer

        raise Exception("Species not pertaining to any buffer.")

    def has_buffer_type(self, name):
        """Is the buffer type identified by ``name`` in this collection?
        """

        for buffer in self.buffer_types:
            if buffer.name == name:
                return True

        return False

    def add_buffer_types(self, buffer_data):
        """Add these buffers to the collection, along with their species.

        :param buffer_data: Dictionary of buffer descriptions, indexed by
           buffer name. Example::

            {"MOPS" : {
                    "species": ["C7H15NO4S", "C7H14NO4S-"],
                    "charges": [0, -1],
                    "pKas": [7.184],
                    },
              "Cacodylic acid" : {
                    "species": ["C2H7AsO2", "C2H6AsO2-"],
                    "charges": [0, -1],
                    "pKas" : [6.28],
              },
              "Acetate" : {
                    "species": ["C2H4O2", "C2H3O2-"],
                    "charges": [0, -1],
                    "pKas" : [4.756],
              },
              "Phosphate" : {
                    "species": ["H3PO4", "H2PO4-", "HPO4-2", "PO4-3"],
                    "charges": [0, -1, -2, -3],
                    "pKas" : [2.148, 7.198, 12.35],
              },
            }

        """

        for buffer_name in buffer_data.keys():
            assert not self.has_buffer_type(buffer_name)
    
            # create species
            species_list = []
    
            for (species_formula, species_charge)\
                    in itertools.imap(None,
                                      buffer_data[buffer_name]["species"],
                                      buffer_data[buffer_name]["charges"]):
                species = SpeciesType(species_formula, species_charge)
                self.add_species(species)
                species_list.append(species)
    
            buffer = BufferType(buffer_name, species_list,
                                buffer_data[buffer_name]["pKas"])
            self.buffer_types.append(buffer)

    def is_buffer_species(self, species):
        """Is ``species`` part of the species part of a buffer type in this 
        collection?
        """

        for buffer in self.buffer_types:
            if species in buffer.species:
                return True

        return False


    def add_species(self, species):
        """Add the given species to this collection.

        :param species: The species to add.
        :type species: :py:class:`SpeciesType`

        :raises: Exception if the species is already in this collection.
        """

        assert not self.has_species(species.formula)

        self.species_types.append(species)

    def add_species_set(self, species_data):
        """Add the set of species in ``species_data`` to this collection.

        :param species_data: A :py:class:`dict` indexed by species' formula,
           whose values are the integer charges associated to each species.

        :raises: Exception if any of the species is already in this collection.
        """

        for (species_formula, species_charge) in species_data.items():
            species = SpeciesType(species_formula, species_charge)
            self.add_species(species)
            

    def get_species_by_formula(self, formula):
        """Get the :py:class:`SpeciesType` identified by the given ``formula``
        in this collection.

        :raises: Exception if no species of that ``formula`` is found in 
          this collection.
        """

        for species in self.species_types:
            if species.formula == formula:
                return species

        raise Exception("No species found for the given formula (%s)."\
                            % formula)
    
    def has_species(self, formula):
        """Is the species identified by ``formula`` in this collection?
        """

        for species in self.species_types:
            if species.formula == formula:
                return True

        return False

    def is_buffer_species_by_formula(self, formula):
        """Is the species identified by ``formula`` part of a buffer type in
        this collection?
        """

        return self.is_buffer_species(self.get_species_by_formula(formula))

    def get_buffer_of_species(self, species):
        """Get the :py:class:`BufferType` in this collection which has
        ``species`` as part of it.

        :raises: Exception if no buffer type in this collection has ``species``.
        """

        for buffer in self.buffer_types:
            if species in buffer.species:
                return buffer

        raise Exception("No buffer has the given species as a component.")

    def get_buffer_of_species_by_formula(self, formula):
        """Get the :py:class:`BufferType` in this collection which has
        the species identified by ``formula`` as part of it.

        :raises: Exception if no buffer type in this collection has ``species``.
        """

        return self.get_buffer_of_species(self.get_species_by_formula(formula))


class SolutionBuffer:
    """A buffer in solution.
    """

    def __init__(self, buffer_type):
        self.buffer_type = buffer_type
        "The type of this buffer. Type: :py:class:`BufferType`."

        self.concentration = None
        """The concentration of this buffer in solution, in Molar units.
        If we know no information, it is ``None``. If we know the concentration
        is unknown, it is ``np.NaN``. All other values are represented as
        float.
        """

class SolutionSpecies:
    """A molecular species in solution.
    """

    def __init__(self, species):
        self.species_type = species
        "The type of this species. Type: :py:class:`SpeciesType`."

        self.concentration = None
        """The concentration of this species in solution, in Molar units.
        If we know no information, it is ``None``. If we know the concentration
        is unknown, it is ``np.NaN``. All other values are represented as
        float.
        """
        

class Solution:
    def __init__(self, buffer_and_species_types, pH):
        """
        :param buffer_and_species_types: The allowed collection of buffer and
          species types for the elements of this solution.
        :type buffer_and_species_types: :py:class:`BufferAndSpeciesTypesDataCollection`.
        :param pH: The final pH of this solution.
        :type pH: float.
        """

        self.species = []
        """List of all the species in solution. Objects are of type
        :py:class:`SolutionSpecies`. The species include salt ions,
        hydronium and hydroxide ions, and the species forming each buffer type.
        There can not be two species of the same :py:class:`SpeciesType`
        in this list.
        """

        self.buffers = []
        """List of all the buffers in solution. Objects are of type
        :py:class:`SolutionBuffer`. Each species forming the buffer is also
        in ``self.species``. There can not be two buffers of the same
        :py:class:`BufferType` in this list.
        """


        self.concentration_data = {} 
        """Maps each buffer in ``self.buffers`` and each species in
        ``self.species`` which is not a buffer species nor hydroxide or
        hydronium ion, to a list of concentrations of that buffer or species
        that is known to be present in the solution. If any of
        the species or buffer is added to the solution in an unknown amount,
        that is also added to the list of concentrations as ``np.NaN``.
        All amounts are in Molar units.

        Example: if we have 0.1 M acetic acid in the solution, and also
        0.1 M sodium acetate, then we have::

           {sodium_ion_species_type : [0.1],
            acetate_buffer_species_type : [0.1, 0.1]}

        Example: if we have 0.1 M acetic acid in the solution, and we 
        add 0.01 M sodium chloride, and then titrate
        with an unknown amount of sodium hydroxide (all molarities
        referred to the final solution), we have::

          {sodium_ion_species_type : [0.01, np.NaN],
           chloride_ion_species_type : [0.01],
           acetate_buffer_species_type : [0.1]}
        
        """

        self.pH = pH
        self.buffer_and_species_types = buffer_and_species_types
        self.concentrations_valid = False

        # add water ions
        h_species = SolutionSpecies(self.buffer_and_species_types.get_species_by_formula("H3O+"))
        oh_species = SolutionSpecies(self.buffer_and_species_types.get_species_by_formula("OH-"))
        
        self.species.append(h_species)
        self.species.append(oh_species)

    def has_species(self, formula):
        """Does this solution contain the species identified by ``formula``?
        """

        for species in self.species:
            if species.species_type.formula == formula:
                return True

        return False

    def has_buffer(self, name):
        """Does this solution contain the buffer identified by ``name``?
        """

        for buffer in self.buffers:
            if buffer.buffer_type.name == name:
                return True

        return False

    def get_buffer_by_name(self, buffer_name):
        """Get the :py:class:`SolutionBuffer` in this solution, identified by
        ``buffer_name``.

        :raises: Exception if this solution does not contain the buffer of
          ``buffer_name``.
        """

        for buffer in self.buffers:
            if buffer.buffer_type.name == buffer_name:
                return buffer
    
        raise Exception("Solution does not have any buffer of the given type (%s)."\
                            % buffer_name)

    def get_species_by_formula(self, formula):
        """Get the :py:class:`SolutionSpecies` in this solution, identified by
        ``formula``.

        :raises: Exception if this solution does not contain the species 
          identified by ``formula``.
        """

        for species in self.species:
            if species.species_type.formula == formula:
                return species

        raise Exception("No solution species with the given formula (%s)."\
                            % formula)

    def get_species(self, species_type):
        """Get the :py:class:`SolutionSpecies` in this solution of type
        ``species_type``.

        :param species_type: The species type required.
        :type species_type: :py:class:`SpeciesType`

        :raises: Exception if this solution does not contain the required
          species.
        """

        for species in self.species:
            if species.species_type == species_type:
                return species

        raise Exception("This species (%s) is not in this solution."\
                            % species_type.formula)

    def add_non_buffer_species(self, formula, concentration):
        """Add a species, identified by ``formula`` to this solution,
        at the given ``concentration``. The species is not part of any buffer.

        :param formula: The formula identifying the species. It cannot be
          any of the water ions (``OH-``, ``H3O+``, ``H+``). It must be one
          of the species allowed for this solution, as stated when creating
          this object (``buffer_and_species_types`` parameter).
        :type formula: :py:func:`str`
        :param concentration: The molar concentration of the species in this solution.
          If it is unknown, the value must be ``NaN``. It adds to the 
          concentration already present in solution.
        :type concentration: :py:func:`float`
        """

        assert formula != "OH-"
        assert formula != "H3O+"
        assert formula != "H+"

        if self.has_species(formula):
            species = self.get_species_by_formula(formula)
            self.concentration_data[species].append(concentration)
        else:
            species_type = self.buffer_and_species_types.get_species_by_formula(formula)
            species = SolutionSpecies(species_type)
            self.species.append(species)
            self.concentration_data[species] = [concentration]

        self.concentrations_valid = False

    def add_buffer(self, name, concentration):
        """Add the buffer, identified by ``name`` to this solution, at the
        given ``concentration``.

        :param name: The name identifying the buffer to add. It must be one
          of the buffers allowed for this solution, as stated when creating
          this object (``buffer_and_species_types`` parameter).
        :type name: :py:func:`str`
        :param concentration: The molar concentration of the buffer (all of its
          species) in this solution. If it is unknown, the value must be `NaN`.
          It adds to the concentration already present in solution.
        :type concentration: py:func:`float`
        """

        if self.has_buffer(name):
            buffer = self.get_buffer_by_name(name)
            self.concentration_data[buffer].append(concentration)
        else:
            buffer_type = self.buffer_and_species_types.get_buffer_type(name)
            buffer = SolutionBuffer(buffer_type)
            self.buffers.append(buffer)
            self.concentration_data[buffer] = [concentration]

            for species_type in buffer_type.species:
                species = SolutionSpecies(species_type)
                assert not self.has_species(species_type.formula)
                self.species.append(species)
                

        self.concentrations_valid = False

    def add_species(self, formula, concentration):
        """Add a species, identified by ``formula`` to this solution,
        at the given ``concentration``. The species may be part of any buffer.

        :param formula: The formula identifying the species. It cannot be
          any of the water ions (``OH-``, ``H3O+``, ``H+``). It must be one
          of the species allowed for this solution, as stated when creating
          this object (``buffer_and_species_types`` parameter).
        :type formula: :py:func:`str`
        :param concentration: The molar concentration of the species in this solution.
          If it is unknown, the value must be ``NaN``. It adds to the 
          concentration already present in solution.
        :type concentration: :py:func:`float`
        """

        if formula in ["H3O+", "H+", "OH-"]:
            pass # ignore
        elif self.buffer_and_species_types.is_buffer_species_by_formula(formula):
            buffer_name = self.buffer_and_species_types.get_buffer_of_species_by_formula(formula).name
            self.add_buffer(buffer_name, concentration)
        else:
            self.add_non_buffer_species(formula, concentration)

    def get_ionic_strength(self):
        """Get the ionic strength of this solution.

        :returns: The ionig strength of this solution, in Molar.
        :rtype: py:func:`float`

        :raises: Exception if the current species in solution, and the
          unknown concentrations are inconsistent.
        """

        if not self.concentrations_valid:
            self.determine_concentrations()

        ionic_strength = 0.0

        for species in self.species:
            charge = species.species_type.charge
            if charge != 0:
                ionic_strength += 0.5 * (charge * charge)\
                    * species.concentration

        return ionic_strength

    def determine_concentrations(self):
        """Find out the unknown concentrations of species in solution.

        :raises: Exception if the current species in solution, and the
          unknown concentrations are inconsistent.
        """

        unknowns = [] # Vector of unknowns. Each element is a SolutionSpecies object.
        unknowns_reverse_index = {} # Indexed by SpeciesType. Returns the unknown index.

        equations_x = []
        equations_y = []

        h_solution_species = self.get_species_by_formula("H3O+")
        oh_solution_species = self.get_species_by_formula("OH-")

        # Set all concentrations as unknown
        for elem in self.species:
            elem.concentration = np.NaN
    
        for buffer in self.buffers:
            buffer.concentration = np.NaN
    
        # 1. Find out H3O+ and OH- concentrations from pH.
        assign_water_ions_concentrations(self.pH, h_solution_species, 
                                         oh_solution_species)

        #   Determine known/unknown concentrations (of species and buffers)
        for (elem, concentrations) in self.concentration_data.items():
            elem.concentration = np.sum(concentrations)

        #   Create unknowns
        for elem in self.species:
            if np.isnan(elem.concentration): # concentration is an unknown
                unknowns.append(elem)

        num_unknowns = len(unknowns)

        if num_unknowns > 0:

            # create reverse index of unknowns
            for i in xrange(num_unknowns):
                unknowns_reverse_index[unknowns[i].species_type] = i

            # 3. For each buffer component A-, create equation
            #    [H+][A-] - 10^(-pKa) * [HA] = 0.0
            for buffer in self.buffers:
                pKa_index = 0 # index to the pKa value corresponding to the studied
                   # acid
                for base_species in buffer.buffer_type.species[1:]:
                    acid_species = buffer.buffer_type.species[pKa_index]
                    pKa = buffer.buffer_type.pKas[pKa_index]

                    new_equation = np.zeros(num_unknowns)

                    # each acid and basis species is unknown because
                    # known concentrations are only for the whole buffer and not
                    # for its species.
                    acid_unknown_index = unknowns_reverse_index[acid_species]
                    base_unknown_index = unknowns_reverse_index[base_species]
                    new_equation[acid_unknown_index] = -math.pow(10, -pKa)
                    new_equation[base_unknown_index] = h_solution_species.concentration

                    equations_x.append(new_equation)
                    equations_y.append(0.0)

                    pKa_index += 1

            # 5. For each buffer of known concentration, create equation
            #    [BC1] + [BC2] + ... + [BCN] = C
            for buffer in self.buffers:
                if not np.isnan(buffer.concentration):
                    new_equation = np.zeros(num_unknowns)
                    for species in buffer.buffer_type.species:
                        species_unknown_index = unknowns_reverse_index[species]
                        new_equation[species_unknown_index] = 1
                    equations_x.append(new_equation)
                    equations_y.append(buffer.concentration)

            # 6. Create charge neutrality equation:
            #    SUM_unknowns([salt_ion]*salt_ion_charge) + SUM_unknowns([buffer_component]*buffer_component_charge) = [H+] - [HO-] - SUM_knowns([salt_ion]*salt_ion_charge) - SUM_knowns([buffer_component]*buffer_component_charge)
            new_equation = np.zeros(num_unknowns)

            for unknown_index in xrange(num_unknowns):
                species = unknowns[unknown_index]
                if species.species_type.charge != 0:
                    new_equation[unknown_index] = species.species_type.charge


            # now, calculate charge sum of all known species
            known_sum = 0.0
            for elem in self.species:
                if elem.species_type.charge != 0:
                    if not np.isnan(elem.concentration):
                        known_sum += elem.concentration * elem.species_type.charge

            equations_x.append(new_equation)
            equations_y.append(-known_sum)

            # 7. Solve set of linear equations with numpy.linalg.lstsq
            x = np.array(equations_x)
            y = np.array(equations_y)

            results = np.linalg.lstsq(x, y)[0]

            # Check constraints
            # - all concentrations are positive
            # - all concentrations are ??
            # Assign concentrations
            for unknown_index in xrange(num_unknowns):
                species = unknowns[unknown_index]
                species.concentration = results[unknown_index]

        # Check no species has an unknown concentration
        # Check all concentrations are positive
        for species in self.species:
            assert not np.isnan(species.concentration),\
                "%s should have a determined value."\
                  % species.species_type.formula
            assert species.concentration >= 0.0,\
                "%s should have a positive concentration (actual concentration is %f)"\
                  % (species.species_type.formula, species.concentration)

        # Assign total buffer concentrations
        for buffer in self.buffers:
            if np.isnan(buffer.concentration):
                concentration = 0.0
                for species_type in buffer.buffer_type.species:
                    concentration += self.get_species(species_type).concentration
                buffer.concentration = concentration

        # Check all concentrations is >= known concentration.
        for (element, concentration_list) in self.concentration_data.items():
            known_values = [item for item in concentration_list if not np.isnan(item)]
            if known_values:
                minimum_concentration = np.sum(known_values)
                assert element.concentration >= minimum_concentration,\
                    "%s should have a concentration above %f (actual value: %f)."\
                      % ((element.species_type.formula if isinstance(element, SolutionSpecies) else element.buffer_type.name),
                         minimum_concentration, element.concentration)

        # Check charge neutrality remains

        self.concentration_valid = True

def assign_water_ions_concentrations(pH, h_solution_species, oh_solution_species):
    """Assign the concentration of hydroxide and hydronium ions in solution.

    :param pH: The solution pH.
    :type pH: :py:func:`float`
    :param h_solution_species: The hydronium species.
    :type h_solution_species: :py:class:`SolutionSpecies`
    :param oh_solution_species: The hydroxide species.
    :type oh_solution_species: :py:class:`SolutionSpecies`

    ``h_solution_species.concentration`` and ``oh_solution_species.concentration``
    are assigned according to the solution pH.
    """

    h_solution_species.concentration = math.pow(10, -pH)
    oh_solution_species.concentration = math.pow(10, -(14 - pH))


def setup_buffer_and_species_types(input_filename=None):
    """Create a :py:class:`BufferAndSpeciesTypesDataCollection` object with the
    species and buffers of ``input_filename``. If it is ``None``, use instead
    the default values (:ref:`sec-default-chem-types-file`).

    :param input_filename: Name of a file following the :ref:`sec-chemical-species-description-file-format`.
    :type input_filename: :py:func:`str`

    :returns: The collection of species and buffers described by the file 
      ``input_filename`` (or the default values).
    :rtype: :py:class:`BufferAndSpeciesTypesDataCollection`
    """

    if not input_filename:
        input_filename = default_buffer_and_species_data_description_filename

    with open(input_filename, 'r') as input:
        data = json.load(input)

    buffer_and_species_types = BufferAndSpeciesTypesDataCollection()
    buffer_and_species_types.add_buffer_types(data["buffers"])
    buffer_and_species_types.add_species_set(data["salts"])
    buffer_and_species_types.add_species_set(data["water"])

    return buffer_and_species_types

def extract_recipe_data(recipe_text):
    """Parse ``recipe_text`` to obtain pairs ``(species, concentration)``,
    where ``species`` is the formula identifying a :py:class:`SpeciesType`,
    and ``concentration`` is the molar concentration of the species.

    :param recipe_text: A :py:func:`str` following the :ref:`sec-solution-description-file-format`.

    :returns: A list of pairs ``(species, concentration)``.
    """

    formatted_recipe = []

    for line in recipe_text.splitlines():
        line = line.strip()

        if (not line) or line.startswith('#'):
            continue

        comment_pos = line.find('#')
        if comment_pos != -1:
            line = line[comment_pos:]

        tokens = line.split(',')
        concentration = float(tokens[-1])
        species = map(lambda x: x.strip(), tokens[:-1])
        formatted_recipe.append( (species, concentration) )

    return formatted_recipe

def extract_recipe_data_from_file(recipe_filename):
    """Parse ``recipe_filename`` to obtain pairs ``(species, concentration)``,
    where ``species`` is the formula identifying a :py:class:`SpeciesType`,
    and ``concentration`` is the molar concentration of the species.

    :param recipe_filename: The name of a file following the :ref:`sec-solution-description-file-format`.

    :returns: A list of pairs ``(species, concentration)``.
    """

    with open(recipe_filename, 'r') as input:
        return extract_recipe_data(input.read())
            
def create_solution_from_recipe(element_types, pH, recipe):
    """Create a :py:class:`Solution` object describing a solution at the given
    ``pH``, following the description of ``recipe``.

    :param element_types: The allowed species and buffers for the solution.
    :type element_types: :py:class:`BufferAndSpeciesTypesDataCollection`
    :param pH: The pH of the solution.
    :type pH: :py:func:`float`
    :param recipe: The species (or buffer species) and concentrations in solution.
    :type recipe: A list of pairs ``(species, concentration)``.

    :returns: A :py:class:`Solution` object as described by this function 
      parameters.
    """

    solution = Solution(element_types, pH=pH)

    for (list_of_species, concentration) in recipe:
        for species in list_of_species:
            solution.add_species(species, concentration)
    
    return solution
