# Copyright 2013 Jorge Estrada. 
#
# This file is part of azucar.buffer project. 
# It is subject to the license terms in the 
# LICENSE file found in the top-level directory of this distribution. 
# No part of azucar.buffer project, including this file, may be copied,
# modified, propagated, or distributed except according to the terms contained 
# in the LICENSE file.

import unittest
import numpy as np

import azucar.buffer


class IS1TestCase (unittest.TestCase):

    def setup_buffer_and_species_types(self):
        return azucar.buffer.setup_buffer_and_species_types()
 
    def test_water_ions_concentration_at_pH_7(self):
        h3o_type = azucar.buffer.SpeciesType("H3O+", +1)
        oh_type = azucar.buffer.SpeciesType("OH-", -1)
        h3o = azucar.buffer.SolutionSpecies(h3o_type)
        oh = azucar.buffer.SolutionSpecies(oh_type)
        pH = 7.0

        azucar.buffer.assign_water_ions_concentrations(pH, h3o, oh)

        self.assertAlmostEqual(1e-7, h3o.concentration)
        self.assertAlmostEqual(1e-7, oh.concentration)

    def test_MOPS_is_in_buffer_types(self):
        buffer_and_species_types = self.setup_buffer_and_species_types()

        self.assertTrue(buffer_and_species_types.has_buffer_type("MOPS"))

    def test_acetate_ion_is_a_buffer_element(self):
        buffer_and_species_types = self.setup_buffer_and_species_types()

        self.assertTrue(
            buffer_and_species_types.is_buffer_species_by_formula("C2H3O2-"))

    def test_sodium_is_in_species_types(self):
        buffer_and_species_types = self.setup_buffer_and_species_types()

        self.assertTrue(
            buffer_and_species_types.has_species("Na+"))

    def test_pure_pH_7_solution_has_1eminus7_ionic_strength(self):
        buffer_and_species_types = self.setup_buffer_and_species_types()
        
        solution = azucar.buffer.Solution(buffer_and_species_types, pH=7.0)

        self.assertAlmostEqual(1e-7, solution.get_ionic_strength())

    def test_sodium_chloride_2_molar_has_same_ionic_strength(self):
        buffer_and_species_types = self.setup_buffer_and_species_types()
        
        solution = azucar.buffer.Solution(buffer_and_species_types, pH=7.0)
        solution.add_species("Na+", 2.0)
        solution.add_species("Cl-", 2.0)

        self.assertAlmostEqual(2.0, solution.get_ionic_strength(), places=5)

    def test_sample_acetate_buffer_ionic_strength(self):
        buffer_and_species_types = self.setup_buffer_and_species_types()
        
        solution = azucar.buffer.Solution(buffer_and_species_types, pH=5.0)
        solution.add_buffer("Acetate", 0.01)
        solution.add_species("Na+", np.NaN)

        self.assertAlmostEqual(0.6369 * 0.01, solution.get_ionic_strength(), places=5)

    def test_sample_mops_buffer_ionic_strength(self):
        buffer_and_species_types = self.setup_buffer_and_species_types()
        
        solution = azucar.buffer.Solution(buffer_and_species_types, pH=7.0)
        solution.add_buffer("MOPS", 0.01)
        solution.add_species("Na+", np.NaN)

        self.assertAlmostEqual(0.3956 * 0.01, solution.get_ionic_strength(), places=5)

    def test_sample_sodium_cacodylate_buffer_ionic_strength(self):
        buffer_and_species_types = self.setup_buffer_and_species_types()
        
        solution = azucar.buffer.Solution(buffer_and_species_types, pH=7.0)
        solution.add_buffer("Cacodylic acid", 0.01)
        solution.add_species("Na+", 0.01)
        solution.add_species("Cl-", np.NaN)

        self.assertAlmostEqual(0.01, solution.get_ionic_strength(), places=5)

    def test_sample_phosphate_buffer_ionic_strength(self):
        buffer_and_species_types = self.setup_buffer_and_species_types()
        
        solution = azucar.buffer.Solution(buffer_and_species_types, pH=7.0)
        solution.add_buffer("Phosphate", 0.01)
        solution.add_species("Na+", 0.01)
        solution.add_species("Na+", np.NaN)

        self.assertAlmostEqual(1.776 * 0.01, solution.get_ionic_strength(), places=5)

    def test_sample_phosphate_buffer_with_added_salts_ionic_strength(self):
        buffer_and_species_types = self.setup_buffer_and_species_types()
        
        solution = azucar.buffer.Solution(buffer_and_species_types, pH=7.0)
        solution.add_buffer("Phosphate", 0.01)
        solution.add_species("Na+", np.NaN)
        solution.add_species("Cl-", 0.01)

        self.assertAlmostEqual(1.776 * 0.01 + 0.01, solution.get_ionic_strength(), places=5)

    def test_add_buffer_as_species(self):
        buffer_and_species_types = self.setup_buffer_and_species_types()
        
        solution = azucar.buffer.Solution(buffer_and_species_types, pH=5.0)
        solution.add_species("C2H4O2", 0.01)
        solution.add_species("Na+", np.NaN)

        self.assertAlmostEqual(0.6369 * 0.01, solution.get_ionic_strength(), places=5)

    def test_create_solution_from_recipe_as_in_phosphate_sample(self):
        buffer_and_species_types = self.setup_buffer_and_species_types()

        recipe = [(["Na+", "H2PO4-"], 0.01),
                  (["Na+", "OH-"], np.NaN),
                  (["Na+", "Cl-"], 0.01)]

        solution = azucar.buffer.create_solution_from_recipe(buffer_and_species_types,
                                                   pH=7.0,
                                                   recipe=recipe)

        self.assertAlmostEqual(1.776 * 0.01 + 0.01, solution.get_ionic_strength(), places=5)
